package com.ninegag.android.library.upload.media

import android.content.Context
import android.util.Log

import com.ninegag.android.library.upload.media.processor.BaseProcessor
import com.ninegag.android.library.upload.media.processor.GifProcessor
import com.ninegag.android.library.upload.media.processor.ImageProcessor
import com.ninegag.android.library.upload.media.processor.Mp4Processor
import com.ninegag.android.library.upload.media.validator.BaseMediaValidator
import com.ninegag.android.library.upload.media.validator.DefaultImageValidatorConfig
import com.ninegag.android.library.upload.media.validator.GifValidatorConfig
import com.ninegag.android.library.upload.media.validator.ImageValidator
import com.ninegag.android.library.upload.media.validator.BaseValidatorConfig
import com.ninegag.android.library.upload.media.validator.Mp4Validator
import com.ninegag.android.library.upload.media.validator.Mp4ValidatorConfig
import com.ninegag.android.library.upload.model.MediaMeta
import com.under9.android.lib.controller.SourceFileController

import java.io.File

/**
 * Created by raymond on 9/5/2017.
 */

open class MediaProcessor(private val context: Context,
                          private val mediaValidatorCallback: BaseMediaValidator.MediaValidatorCallback,
                          private val saveMediaCallback: BaseProcessor.SaveMediaCallback,
                          private val shouldCompress: Boolean = true) {

    private var baseProcessor: BaseProcessor? = null
    private var baseMediaValidator: BaseMediaValidator? = null

    fun processMediaMeta(filePath: String, tmpFilePath: String) {
        @MediaMeta.MediaType val mediaType = getMediaType(filePath)
        val config = createValidatorConfigByType(mediaType)

        val f = File(filePath)
        baseMediaValidator = createMediaValidatorByType(mediaType, config, mediaValidatorCallback)
        baseProcessor = createProcessorByType(context, saveMediaCallback, baseMediaValidator, mediaType)

        val mediaMeta = baseProcessor!!.getMediaMeta(f)
        baseProcessor!!.saveMediaToTempFile(mediaMeta, tmpFilePath)
    }

    /**
     * To
     * @param mediaType
     * @return
     */
    protected open fun createValidatorConfigByType(@MediaMeta.MediaType mediaType: Int): BaseValidatorConfig {
        when (mediaType) {
            MediaMeta.MEDIA_TYPE_IMAGE -> return DefaultImageValidatorConfig()
            MediaMeta.MEDIA_TYPE_GIF -> return GifValidatorConfig()
            MediaMeta.MEDIA_TYPE_MP4 -> return Mp4ValidatorConfig()
            else -> return DefaultImageValidatorConfig()
        }
    }

    /**
     *
     * @param mediaType
     * @param config
     * @param callback
     * @return
     */
    protected fun createMediaValidatorByType(@MediaMeta.MediaType mediaType: Int,
                                             config: BaseValidatorConfig,
                                             callback: BaseMediaValidator.MediaValidatorCallback): BaseMediaValidator? {
        when (mediaType) {
            MediaMeta.MEDIA_TYPE_IMAGE, MediaMeta.MEDIA_TYPE_GIF -> return ImageValidator(config, callback)
            MediaMeta.MEDIA_TYPE_MP4 -> return Mp4Validator(config, callback)
            else -> return null
        }
    }

    /**
     *
     *
     * @param context
     * @param validator
     * @param mediaType
     * @return
     */
    protected open fun createProcessorByType(context: Context,
                                             saveMediaCallback: BaseProcessor.SaveMediaCallback,
                                             validator: BaseMediaValidator?, @MediaMeta.MediaType mediaType: Int): BaseProcessor {
        when (mediaType) {
            MediaMeta.MEDIA_TYPE_IMAGE -> return ImageProcessor(saveMediaCallback, validator!!, shouldCompress)
            MediaMeta.MEDIA_TYPE_GIF -> return GifProcessor(saveMediaCallback, validator)
            MediaMeta.MEDIA_TYPE_MP4 -> {
                val sfc = SourceFileController(true)
                sfc.init(context)
                return Mp4Processor(saveMediaCallback, validator, context, sfc)
            }
            else -> return ImageProcessor(saveMediaCallback, validator!!, shouldCompress)
        }
    }

    companion object {
        private val TAG = "MediaProcessor"

        @JvmStatic
        fun getMediaType(filePath: String?): Int {
            // Fallback to image
            if (filePath == null) return MediaMeta.MEDIA_TYPE_IMAGE

            try {
                val index = filePath.lastIndexOf(".")
                val ext = filePath.substring(index, filePath.length)
                when (ext.toLowerCase()) {
                    ".gif" -> return MediaMeta.MEDIA_TYPE_GIF
                    ".mp4" -> return MediaMeta.MEDIA_TYPE_MP4
                    else -> return MediaMeta.MEDIA_TYPE_IMAGE
                }
            } catch (e: StringIndexOutOfBoundsException) {
                Log.e(TAG, "getFileType: ", e)
                return MediaMeta.MEDIA_TYPE_IMAGE
            }

        }
    }
}
