package com.under9.android.comments.adapter

import android.content.Intent
import androidx.annotation.CallSuper
import com.under9.android.comments.CommentSystem
import com.under9.android.comments.controller.CommentSystemTaskQueueController
import com.under9.android.comments.controller.MetricsController
import com.under9.android.comments.data.repository.LocalCommentListRepositoryInterface
import com.under9.android.comments.data.repository.UserRepositoryInterface
import com.under9.android.comments.event.DeleteCommentDoneEvent
import com.under9.android.comments.event.DeleteCommentEvent
import com.under9.android.comments.model.CommentItem
import com.under9.android.comments.model.wrapper.CommentItemWrapperInterface
import com.under9.android.comments.task.TaskCallback
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import timber.log.Timber

/**
 * Base handler that modify [CommentItemWrapperInterface] model
 */
abstract class BaseCommentListItemHandler(
        private val scope: String?,
        private val userRepositoryInterface: UserRepositoryInterface,
        private val localCommentListRepository: LocalCommentListRepositoryInterface,
        private val commentSystemTaskQueueController: CommentSystemTaskQueueController
): BaseCommentListItemAction() {

    override fun onItemMore(position: Int, commentWrapper: CommentItemWrapperInterface, username: String) {
        super.onItemMore(position, commentWrapper, username)
    }

    override fun onItemCopy(commentWrapper: CommentItemWrapperInterface) {
        super.onItemCopy(commentWrapper)
    }

    override fun onItemShare(commentWrapper: CommentItemWrapperInterface) {
        super.onItemShare(commentWrapper)
    }

    override fun onItemDelete(position: Int, commentWrapper: CommentItemWrapperInterface) {
        super.onItemDelete(position, commentWrapper)
    }

    override fun onItemDeleteAccpeted(position: Int, commentWrapper: CommentItemWrapperInterface) {
        super.onItemDeleteAccpeted(position, commentWrapper)
        commentWrapper.apply {
            status = CommentItem.STATUS_DELETED
        }
        updateCommentInDb(localCommentListRepository, commentWrapper)
        MetricsController.logUserActions(MetricsController.USER_ACTIONS_DELETE)
        commentSystemTaskQueueController.issueDelete(commentWrapper.getAttachedToUrl(), commentWrapper.getCommentId(), object: TaskCallback() {
            override fun run(data: Intent?) {
                CommentSystem.postEvent(scope, DeleteCommentDoneEvent(commentWrapper.getCommentId()))
            }
        })
        CommentSystem.postEvent(scope, DeleteCommentEvent(commentWrapper.getCommentId()))
    }

    @CallSuper
    override fun onItemReport(position: Int, commentWrapper: CommentItemWrapperInterface) {
        commentWrapper.apply {
            commentWrapper.status = CommentItem.STATUS_REPORTED
        }
        updateCommentInDb(localCommentListRepository, commentWrapper)
    }

    @CallSuper
    override fun onItemLike(position: Int, commentWrapper: CommentItemWrapperInterface): Boolean {
        return if (commentWrapper.likeStatus != CommentItem.LIKE_STATUS_LIKED) {
            commentWrapper.apply {
                if (likeStatus == CommentItem.LIKE_STATUS_DISLIKED) dislikeCount--
                likeStatus = CommentItem.LIKE_STATUS_LIKED
                likeCount++
            }
            userRepositoryInterface.addLikeOrDislikedComment(commentWrapper.getCommentId(), CommentItem.LIKE_STATUS_LIKED)
            updateCommentInDb(localCommentListRepository, commentWrapper)
            commentSystemTaskQueueController.issueLike(commentWrapper.getCommentId(), null)
            MetricsController.logUserActions(MetricsController.USER_ACTIONS_UPVOTE)
            true
        } else {
            false
        }

    }

    @CallSuper
    override fun onItemUnlike(position: Int, commentWrapper: CommentItemWrapperInterface): Boolean {
        commentWrapper.apply {
            when (commentWrapper.likeStatus) {
                CommentItem.LIKE_STATUS_LIKED -> commentWrapper.likeCount--
                CommentItem.LIKE_STATUS_DISLIKED -> commentWrapper.dislikeCount--
            }
            likeStatus = CommentItem.LIKE_STATUS_NORMAL
        }
        userRepositoryInterface.addLikeOrDislikedComment(commentWrapper.getCommentId(), CommentItem.LIKE_STATUS_NORMAL)
        updateCommentInDb(localCommentListRepository, commentWrapper)
        commentSystemTaskQueueController.issueUnlike(commentWrapper.getCommentId(), null)
        return true
    }

    @CallSuper
    override fun onItemDislike(position: Int, commentWrapper: CommentItemWrapperInterface): Boolean {
        return if (commentWrapper.likeStatus != CommentItem.LIKE_STATUS_DISLIKED) {
            commentWrapper.apply {
                if (likeStatus == CommentItem.LIKE_STATUS_LIKED) likeCount--
                likeStatus = CommentItem.LIKE_STATUS_DISLIKED
                dislikeCount++
            }
            MetricsController.logUserActions(MetricsController.USER_ACTIONS_DOWNVOTE)
            userRepositoryInterface.addLikeOrDislikedComment(commentWrapper.getCommentId(), CommentItem.LIKE_STATUS_DISLIKED)
            updateCommentInDb(localCommentListRepository, commentWrapper)
            commentSystemTaskQueueController.issueDislike(commentWrapper.getCommentId(), null)

            true
        } else {
            false
        }
    }

    private fun updateCommentInDb(
            localCommentListRepository: LocalCommentListRepositoryInterface,
            commentWrapper: CommentItemWrapperInterface) {
        Single.fromCallable {
            val commentItem = localCommentListRepository.getCommentItemById(commentWrapper.getCommentId())
            Timber.d("commentItem=$commentItem, wrapper=$commentWrapper")
            commentItem?.apply {
                likeCount = commentWrapper.likeCount
                dislikeCount = commentWrapper.dislikeCount
                likeStatus = commentWrapper.likeStatus
                status = commentWrapper.status
            }?.also {
                val commentListItems = localCommentListRepository.getCommentListItemByCommentDbId(commentItem.id)
                commentListItems.forEach { commentListItem ->
                    commentListItem.apply {
                        status = commentWrapper.status
                        update()
                    }
                }

                commentItem.update()
                commentItem.refresh()
            }
        }.subscribeOn(Schedulers.io()).subscribe()
    }
}