package com.under9.android.comments.data.repository

import android.net.Uri
import android.util.ArrayMap
import com.ninegag.android.blitz2.BlitzQueryResult
import com.ninegag.android.blitz2.BlitzTransformer
import com.ninegag.android.blitz2.PagingInfo
import com.ninegag.android.blitz2.RxDataSource
import com.under9.android.comments.data.query.comment.CommentListQueryParam2
import com.under9.android.comments.data.query.comment.CommentListQueryResult2
import com.under9.android.comments.data.query.user.UserStatusQueryResult
import com.under9.android.comments.helper.CommentListHelper
import com.under9.android.comments.model.CommentItem
import com.under9.android.comments.model.CommentList
import com.under9.android.comments.model.CommentListItem2
import com.under9.android.comments.model.wrapper.CommentItemWrapper
import com.under9.android.comments.model.wrapper.CommentItemWrapperInterface
import io.reactivex.Observable
import io.reactivex.functions.BiFunction
import timber.log.Timber
import java.util.concurrent.atomic.AtomicInteger

class CommentDataSource(
        private val localCommentListRepository: LocalCommentListRepositoryInterface,
        private val commentListRepository: CommentListRepositoryInterface,
        private val userRepository: UserRepositoryInterface
) : RxDataSource<CommentItemWrapperInterface, String, CommentListQueryParam2>() {
    val TAG = "CommentListWrapper"

    var url: String? = null
    var commentChildrenUrl: String? = null
    var commentId: String? = null
    var loadType: Int = CommentListQueryParam2.LOAD_TYPE_HOT
    var direction: Int = CommentListQueryParam2.DIRECTION_NEXT

    /**
     * As an identifier for this datasource
     */
    var listKey: String? = null
        private set(value) { field = value }
    private var lastRemoteRefreshTs: Long = 0

    /**
     * Calculate what will be the next offset when fetching in database
     */
    private var localNextOffset = AtomicInteger(0)

    private val transformer: BlitzTransformer<CommentItem, CommentItemWrapperInterface> = object: BlitzTransformer<CommentItem, CommentItemWrapperInterface> {
        override fun transform(list: List<CommentItem>): List<CommentItemWrapperInterface> {
            val itemList = arrayListOf<CommentItemWrapperInterface>()
            list.forEachIndexed { _, item ->
                val wrapper = CommentItemWrapper.obtainInstance(item, userRepository.getLoggedInUserId())
                val likeStatus = userRepository.getCommentLikeStatus(item.commentId)
                Timber.d("commentId=${item.commentId}, text=${wrapper.getContent()}, likeStatus=$likeStatus")
                wrapper.likeStatus = userRepository.getCommentLikeStatus(item.commentId)
                itemList.add(wrapper)
            }
            return itemList
        }
    }

    override fun createInitLoadObservable(queryParam: CommentListQueryParam2): Observable<BlitzQueryResult<CommentItemWrapperInterface, String>> {
        return Observable.defer {
            Observable.just(localCommentListRepository.obtainListByListKey(queryParam.listKey))
                    .zipWith(Observable.just(localCommentListRepository.getNextListItems(queryParam.listKey, 0, CommentListQueryParam2.DEFAULT_COUNT)), BiFunction<CommentList, List<CommentListItem2>, CommentListQueryResult2> { commentList, commentListItem ->
                        this.lastRemoteRefreshTs = commentList.lastFetch ?: 0
                        CommentListQueryResult2(commentList.prevCursor, commentList.nextCursor, commentList.opUserId, commentList.level, commentListItem)
                    })
                    .map {
                        Timber.tag(TAG).v("local=${it.comments.size}")
                        val itemList = ArrayList<CommentItem>()
                        it.comments.forEach { commentListItem ->
                            commentListItem.commentItem.isOP = (commentListItem.commentItem.user != null && commentListItem.commentItem.user.userId == it.opUserId)
                            commentListItem.commentItem.prevUrl = it.prev

                            Timber.d("initLoad isOP=${commentListItem.commentItem.isOP}, opUserId=${it.opUserId}, userId=${commentListItem.commentItem?.user?.userId}")
                            itemList.add(commentListItem.commentItem)
                        }
                        transformer.transform(itemList)
                    }.flatMap {
                        if (queryParam.isRefresh || it.isEmpty()) {
                            val zipperFunction = BiFunction<UserStatusQueryResult, CommentListQueryResult2, CommentListQueryResult2> { userStatus, commentListResult ->
                                commentListResult
                            }
                            userRepository.mayFetchUserStatus().zipWith(commentListRepository.getCommentList2(queryParam), zipperFunction).toObservable().flatMap { commentListQueryResult ->
                                Timber.tag(TAG).v("remote=${commentListQueryResult.comments.size}, $url")
                                val itemList = ArrayList<CommentItem>()
                                commentListQueryResult.comments.forEach { commentListItem ->
                                    commentListItem.commentItem.isOP = (commentListItem.commentItem.user != null && commentListItem.commentItem.user.userId == commentListQueryResult.opUserId)
                                    commentListItem.commentItem.prevUrl = commentListQueryResult.prev
                                    itemList.add(commentListItem.commentItem)
                                }

                                val formattedPrevKey = commentListQueryResult.prev
                                val formattedNextKey = commentListQueryResult.next
                                
                                localNextOffset.set(itemList.size)

                                val metaMap = ArrayMap<String, String>()
                                metaMap[CommentListQueryResult2.KEY_LEVEL] = commentListQueryResult.level.toString()
                                metaMap[CommentListQueryResult2.KEY_OP_USER_ID] = commentListQueryResult.opUserId
                                Observable.just(BlitzQueryResult(
                                        items = transformer.transform(itemList),
                                        prevKey = formattedPrevKey,
                                        nextKey = formattedNextKey,
                                        listMeta = metaMap
                                ))
                            }
                        } else {
                            val cursor = localCommentListRepository.getListMeta(queryParam.listKey, queryParam.direction)
                            val metaMap = ArrayMap<String, String>()
                            metaMap[CommentListQueryResult2.KEY_LEVEL] = cursor[CommentListQueryResult2.KEY_LEVEL]
                            metaMap[CommentListQueryResult2.KEY_OP_USER_ID] = cursor[CommentListQueryResult2.KEY_OP_USER_ID]

                            localNextOffset.set(it.size)

                            val formattedPrevKey = cursor[CommentListQueryResult2.KEY_PREV]
                            val formattedNextKey = cursor[CommentListQueryResult2.KEY_NEXT]

                            Observable.just(BlitzQueryResult(
                                    items = it,
                                    prevKey = formattedPrevKey,
                                    nextKey = formattedNextKey,
                                    listMeta = metaMap
                            ))
                        }
                    }
        }
    }

    override fun createLoadNextObservable(queryParam: CommentListQueryParam2): Observable<BlitzQueryResult<CommentItemWrapperInterface, String>> {
        return commentListRepository.getCommentList2(queryParam).toObservable()
                .flatMap { commentListQueryResult ->
                    Timber.tag(TAG).v("loadNext, queryParam=$queryParam")

                    Timber.tag(TAG).v("remote=${commentListQueryResult.comments.size}, $url")
                    val itemList = ArrayList<CommentItem>()
                    commentListQueryResult.comments.forEach { commentListItem ->
                        commentListItem.commentItem.isOP = (commentListItem.commentItem.user != null && commentListItem.commentItem.user.userId == commentListQueryResult.opUserId)
                        commentListItem.commentItem.prevUrl = commentListQueryResult.prev

                        itemList.add(commentListItem.commentItem)
                    }

                    localNextOffset.set(queryParam.localNextOffset + itemList.size)

                    val formattedPrevKey = commentListQueryResult.prev
                    val formattedNextKey = commentListQueryResult.next

                    Observable.just(BlitzQueryResult(
                            items = transformer.transform(itemList),
                            prevKey = formattedPrevKey,
                            nextKey = formattedNextKey
                    ))
                }
    }

    override fun createLoadPrevObservable(queryParam: CommentListQueryParam2): Observable<BlitzQueryResult<CommentItemWrapperInterface, String>> {
        return commentListRepository.getCommentList2(queryParam).toObservable()
                .flatMap { commentListQueryResult ->
                    Timber.tag(TAG).v("loadPrev, queryParam=$queryParam")

                    Timber.tag(TAG).v("remote=${commentListQueryResult.comments.size}, $url")
                    val itemList = ArrayList<CommentItem>()
                    commentListQueryResult.comments.forEach { commentListItem ->
                        commentListItem.commentItem.isOP = (commentListItem.commentItem.user != null && commentListItem.commentItem.user.userId == commentListQueryResult.opUserId)
                        itemList.add(commentListItem.commentItem)
                    }

                    val formattedPrevKey = commentListQueryResult.prev
                    val formattedNextKey = commentListQueryResult.next

                    localNextOffset.set(queryParam.localNextOffset + itemList.size)

                    Observable.just(BlitzQueryResult(
                            items = transformer.transform(itemList),
                            prevKey = formattedPrevKey,
                            nextKey = formattedNextKey,
                            prependOffset = commentListQueryResult.level - 1
                    ))
                }
    }

    override fun createQueryParam(pageInfo: PagingInfo<CommentItemWrapperInterface, String>?): CommentListQueryParam2 {
        return if (pageInfo != null) {
            // Has pages
            val prevKey = pageInfo.queryResult.prevKey ?: ""
            val nextKey = pageInfo.queryResult.nextKey ?: ""

            if (direction == CommentListQueryParam2.DIRECTION_NEXT) {
                createQueryParamWithKey(nextKey, localNextOffset.get())
            } else {
                createQueryParamWithKey(prevKey, localNextOffset.get())
            }
        } else {
            // First load
            createInvalidationQueryParam()
        }
    }

    override fun createInvalidationQueryParam(): CommentListQueryParam2 {
        val param = CommentListQueryParamFactory.createInitCommentListQueryParam(url!!, commentId, commentChildrenUrl, prevKey, nextKey, loadType)
        this.listKey = param.listKey
        this.localNextOffset.set(0)
        return param
    }

    override fun hasNext(pagingInfo: PagingInfo<CommentItemWrapperInterface, String>): Boolean {
        return pagingInfo.queryResult.nextKey != null
    }

    override fun hasPrev(pagingInfo: PagingInfo<CommentItemWrapperInterface, String>): Boolean {
        return pagingInfo.queryResult.prevKey != null
    }

    private fun createQueryParamWithKey(key: String, localOffset: Int): CommentListQueryParam2 {
        val uri = Uri.parse("$url?$key")
        val param = CommentListQueryParam2(
                isRefresh = false,
                direction = direction,
                listKey = listKey!!,
                url = uri.getQueryParameter(CommentListQueryParam2.KEY_URL),
                type = uri.getQueryParameter(CommentListQueryParam2.KEY_TYPE) ?: CommentListQueryParam2.DEFAULT_TYPE,
                count = uri.getQueryParameter(CommentListQueryParam2.KEY_COUNT)?.toInt() ?: CommentListQueryParam2.DEFAULT_COUNT,
                commentId = (uri.getQueryParameter(CommentListQueryParam2.KEY_COMMENT_ID) ?: null),
                level = (uri.getQueryParameter(CommentListQueryParam2.KEY_LEVEL) ?: null)?.toInt(),
                prev = uri.getQueryParameter(CommentListQueryParam2.KEY_PREV),
                next = uri.getQueryParameter(CommentListQueryParam2.KEY_NEXT),
                localNextOffset = localOffset
        )

        Timber.tag(TAG).v("createQueryParamWithKey=$param, uri=$uri")
        return param
    }

    /**
     * Factory method to create different loading params
     */
    object CommentListQueryParamFactory {
        @JvmStatic fun createInitCommentListQueryParam(url: String,
                                                       commentId: String?,
                                                       commentChildrenUrl: String?,
                                                       prev: String?, next: String?, loadType: Int): CommentListQueryParam2 {
            return when (loadType) {

                CommentListQueryParam2.LOAD_TYPE_RESTORE_POSITION_FIRST_LEVEL -> {
                    CommentListQueryParam2(
                            true, CommentListQueryParam2.DIRECTION_NEXT,
                            CommentListHelper.formatListKey(url, null, null),
                            url,
                            CommentListQueryParam2.TYPE_OLD,
                            CommentListQueryParam2.DEFAULT_COUNT,
                            commentId, null, null, null, 0)
                }

                CommentListQueryParam2.LOAD_TYPE_THREAD -> {
                    val uri = Uri.parse("$url?$commentChildrenUrl")

                    val urlCommentId = uri.getQueryParameter(CommentListQueryParam2.KEY_COMMENT_ID) ?: commentId
                    val level = uri.getQueryParameter(CommentListQueryParam2.KEY_LEVEL)?.toInt() ?: 2
                    val count = uri.getQueryParameter(CommentListQueryParam2.KEY_COUNT)?.toInt() ?: CommentListQueryParam2.DEFAULT_COUNT

                    CommentListQueryParam2(
                            isRefresh = true,
                            direction = CommentListQueryParam2.DIRECTION_NEXT,
                            listKey = CommentListHelper.formatListKey(url, commentId, level),
                            url = url,
                            type = uri.getQueryParameter(CommentListQueryParam2.KEY_TYPE),
                            count = count,
                            commentId = urlCommentId,
                            level = level,
                            prev = uri.getQueryParameter(CommentListQueryParam2.KEY_PREV),
                            next = uri.getQueryParameter(CommentListQueryParam2.KEY_NEXT),
                            localNextOffset = 0
                    )
                }

                CommentListQueryParam2.LOAD_TYPE_LINKED_COMMENT -> {
                    CommentListQueryParam2(
                            true, CommentListQueryParam2.DIRECTION_NEXT,
                            CommentListHelper.formatListKey(url, commentId, 2),
                            url,
                            CommentListQueryParam2.DEFAULT_TYPE,
                            CommentListQueryParam2.DEFAULT_COUNT,
                            commentId, 2, null, null, 0)
                }

                CommentListQueryParam2.LOAD_TYPE_NEW -> {
                    CommentListQueryParam2(
                            true, CommentListQueryParam2.DIRECTION_NEXT,
                            CommentListHelper.formatListKey(url, null, 1),
                            url,
                            CommentListQueryParam2.TYPE_NEW,
                            CommentListQueryParam2.DEFAULT_COUNT,
                            null, 1, null, null, 0)
                }
                
                else -> CommentListQueryParam2(
                        true, CommentListQueryParam2.DIRECTION_NEXT,
                        CommentListHelper.formatListKey(url, null, 1),
                        url,
                        CommentListQueryParam2.DEFAULT_TYPE,
                        CommentListQueryParam2.DEFAULT_COUNT,
                        null, 1, null, null, 0)
            }
        }
    }
}