package com.under9.android.comments.data.repository

import com.under9.android.comments.api.CommentApiService
import com.under9.android.comments.controller.DataController
import com.under9.android.comments.data.query.comment.CommentListQueryParam
import com.under9.android.comments.data.query.comment.CommentListQueryParam2
import com.under9.android.comments.data.query.comment.CommentListQueryResult
import com.under9.android.comments.data.query.comment.CommentListQueryResult2
import com.under9.android.comments.data.processor.CommentListProcessor
import com.under9.android.comments.data.processor.CommentListProcessor2
import com.under9.android.comments.helper.CommentHelper
import com.under9.android.comments.helper.CommentWriterV2
import com.under9.android.lib.network.Transformations
import io.reactivex.Flowable
import io.reactivex.functions.Function
import org.reactivestreams.Publisher
import timber.log.Timber

class CommentListRepository(dataController: DataController,
                            private val api: CommentApiService,
                            private val localUserRepository: LocalUserRepositoryInterface,
                            private val localCommentListRepository: LocalCommentListRepositoryInterface
): BaseRepository(dataController), CommentListRepositoryInterface {

    override fun getCommentList(commentListQueryParam: CommentListQueryParam): Flowable<CommentListQueryResult> {
        commentListQueryParam.run {
            return api.getCommentList(dataController.appId,
                    url,
                    CommentHelper.getOrderString(order),
                    count,
                    ref, refCommentId, direction, commentId)
                    .compose(Transformations.flowableResultToResponseWithWithHttpErrorHandling(5))
                    .flatMap {
                        val apiCommentList = it.body()

                        apiCommentList?.payload?.comments?.forEach { comment ->
                            Timber.d("api, subLevelLoadMore=$comment")
                        }


                        if (apiCommentList == null) {
                            Flowable.just(CommentListQueryResult(false, ""))
                        } else {
                            val processor = CommentListProcessor(dataController, commentListQueryParam, CommentWriterV2.instance)
                            processor.process(apiCommentList)
                        }
                    }
                    .onErrorResumeNext(Function<Throwable, Publisher<CommentListQueryResult>> {
                        Timber.e(it, "onErrorResumeNext")
                        Flowable.just(CommentListQueryResult(true, dataController.getLastOrderKey(commentListQueryParam.url, commentListQueryParam.order)))
                    })
        }
    }

    override fun getCommentList2(commentListQueryParam: CommentListQueryParam2): Flowable<CommentListQueryResult2> {
        commentListQueryParam.run {
            return api.getCommentList2(dataController.appId, url, count, commentId, type, level, prev, next)
                    .doOnNext {
                        Timber.tag("CommentListWrapper").v("in repo url=${it.response()?.raw()?.request()?.url()}, queryParam=$commentListQueryParam")
                    }
                    .compose(Transformations.flowableResultToResponseWithWithHttpErrorHandling(1))
                    .flatMap {
                        val processor = CommentListProcessor2(dataController, commentListQueryParam, localUserRepository, localCommentListRepository)
                        processor.process(it.body()!!)
                    }
                    .onErrorResumeNext(Function<Throwable, Publisher<CommentListQueryResult2>> {
                        Timber.w(it, "onErrorResumeNext")

                        val items = localCommentListRepository.getNextListItems(listKey, localNextOffset, 300)
                        val listMeta = localCommentListRepository.getListMeta(listKey, direction)

                        Flowable.just(CommentListQueryResult2(
                                listMeta[CommentListQueryResult2.KEY_PREV],
                                listMeta[CommentListQueryResult2.KEY_NEXT],
                                listMeta[CommentListQueryResult2.KEY_OP_USER_ID] ?: error(""),
                                (listMeta[CommentListQueryResult2.KEY_LEVEL] ?: error("")).toInt(), items
                        ))
                    })

        }
    }


}