package com.under9.android.comments.model.wrapper

import android.net.Uri
import android.text.Html
import android.text.Spanned
import com.ninegag.android.blitz.BlitzItem
import com.under9.android.comments.CommentSystem
import com.under9.android.comments.data.query.comment.CommentListQueryParam2
import com.under9.android.comments.helper.CommentListHelper
import com.under9.android.comments.model.CommentItem
import com.under9.android.comments.model.api.MediaData
import com.under9.android.comments.model.constant.CommentConstant
import timber.log.Timber
import java.lang.RuntimeException
import java.lang.ref.WeakReference
import java.util.*

class CommentItemWrapper private constructor(
        underlyingObject: CommentItem
): BlitzItem<CommentItem>(underlyingObject), CommentItemWrapperInterface {

    companion object {
        private const val NEW_INTERVAL = 5000
        private const val FORCE_HIGHLIGHT_INTERVAL = NEW_INTERVAL

        private val objectMap = WeakHashMap<CommentItem, WeakReference<CommentItemWrapper>>()

        @JvmStatic
        fun obtainInstance(commentItem: CommentItem, myUserId: String?): CommentItemWrapperInterface {
            synchronized(objectMap) {
                val ref = objectMap[commentItem]
                if (ref != null) {
                    val obj = ref.get()
                    if (obj != null) {
                        obj.underlyingObject = commentItem
                        obj.myUserId = myUserId
                        return obj
                    }
                }

                val obj = CommentItemWrapper(commentItem)
                objectMap[commentItem] = WeakReference(obj)
                obj.myUserId = myUserId
                return obj
            }
        }
    }

    private var myUserId: String? = null
    private var cachedContent: Spanned? = null

    override var isOP: Boolean = false
        get() {
            Timber.d("underlyingObject.isOP=${underlyingObject.isOP}")
            return underlyingObject.isOP
        }
        set(value) {
            field = value
            underlyingObject.isOP = value
            Timber.d("setOP=$value")
        }
    override var likeStatus: Int = underlyingObject.likeStatus ?: 0

    override fun getId(): Long = underlyingObject.id

    override fun getCommentId(): String = underlyingObject.commentId

    override var likeCount: Int = underlyingObject.likeCount
    override var dislikeCount: Int = underlyingObject.dislikeCount

    override var status: Int = CommentItem.STATUS_NORMAL
        get() = underlyingObject.status ?: CommentItem.STATUS_NORMAL
        set(value) {
            field = value
            underlyingObject.status = value
        }

    override fun shouldHighlight(): Boolean {
        val curr = System.currentTimeMillis()
        val diff = curr - getTime()
        Timber.d("shouldHighlight=$diff, cur=$curr, obj=${underlyingObject.timestamp}, isPending=${isPending()}")
        return isPending() || (diff < NEW_INTERVAL && diff > 0)
    }

    override fun isPending(): Boolean {
        return underlyingObject.status == CommentItem.STATUS_PENDING
    }

    override fun getMentionedAccountId(username: String): String? = underlyingObject.mentionMap[username]

    override fun getMentionedUserMap(): Map<String, String> = underlyingObject.mentionMap

    override fun getPermalink(): String = underlyingObject.permalink

    override fun getShareUrl(): String = getPermalink().replace("http", "https") + "?ref=android"

    override fun getLevel(): Int = underlyingObject.level

    override fun getContent(): Spanned? {
        if (cachedContent == null) {
            val text = if (underlyingObject.mediaList.isNullOrEmpty()) underlyingObject.text else underlyingObject.mediaText
            cachedContent = Html.fromHtml(text)
        }
        return cachedContent
    }

    override fun getRawText(): String? = underlyingObject.text

    override fun getCommentLabel(): String {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }

    override fun getType(): Int {
        return when (underlyingObject.type) {
            CommentItem.MEDIA_TYPE_MEDIA, CommentItem.MEDIA_TYPE_USER_MEDIA -> CommentConstant.TYPE_MEDIA
            CommentItem.MEDIA_TYPE_TEXT -> CommentConstant.TYPE_TEXT
            else -> throw RuntimeException("Unknown comment type")
        }
    }

    override fun getUser(): UserItemWrapperInterface {
        return if (underlyingObject.user != null) {
            UserWrapper.obtainInstance(underlyingObject.user)
        } else {
            PendingUserWrapper(underlyingObject.user, CommentSystem.getInstance().context)
        }
    }

    override fun getTime(): Long = underlyingObject.timestamp * 1000

    override fun getChildrenUrl(): String? {
        return underlyingObject.childrenUrl
    }

    override fun getChildCommentListQueryParam(): CommentListQueryParam2? {
        return if (underlyingObject.childrenUrl != null) {
            val uri = Uri.parse("${underlyingObject.url}?${underlyingObject.childrenUrl}")
            val commentId = uri.getQueryParameter(CommentListQueryParam2.KEY_COMMENT_ID)!!
            val level = uri.getQueryParameter(CommentListQueryParam2.KEY_LEVEL)?.toInt() ?: 1
            val count = uri.getQueryParameter(CommentListQueryParam2.KEY_COUNT)?.toInt() ?: CommentListQueryParam2.DEFAULT_COUNT

            CommentListQueryParam2(
                    isRefresh = true,
                    direction = CommentListQueryParam2.DIRECTION_NEXT,
                    listKey = CommentListHelper.formatListKey(underlyingObject.url, commentId, level),
                    url = underlyingObject.url,
                    type = null,
                    count = count,
                    commentId = commentId,
                    level = null,
                    prev = null,
                    next = null,
                    localNextOffset = 0
            )
        } else null
    }

    override fun getMediaList(): Array<MediaData>? = underlyingObject.mediaList

    override fun getFirstMedia(): MediaData? {
        return if (underlyingObject.mediaList.isNullOrEmpty()) {
            null
        } else {
            underlyingObject.mediaList!![0]
        }
    }

    override fun getChildrenTotal(): Int = underlyingObject.childrenTotal

    override fun isMyComment(): Boolean {
        if (myUserId == null) return false
        if (underlyingObject.user == null) return false

        Timber.d("isMyComment=$myUserId, commentUserId${underlyingObject.user.userId}, content=${underlyingObject.mediaText}")

        return underlyingObject.user.userId == myUserId || underlyingObject.user.userId == "anon_$myUserId"
    }

    override fun getAttachedToUrl(): String = underlyingObject.url

    override fun isParent(): Boolean = getLevel() == 1

    override var prevUrl: String?
        get() = underlyingObject.prevUrl
        set(value) {
            underlyingObject.prevUrl = value
        }

    override fun refresh() {
        cachedContent = null
        underlyingObject.invalidate()
        underlyingObject.refresh()
    }

    override fun toString(): String {
        return "title=${underlyingObject.text}, " + super.toString()
    }
}