package com.under9.android.comments.ui.renderer

import android.animation.ObjectAnimator
import android.graphics.drawable.ColorDrawable
import android.graphics.drawable.TransitionDrawable
import android.net.Uri
import android.os.Bundle
import android.text.Spannable
import android.text.SpannableStringBuilder
import android.text.style.ForegroundColorSpan
import android.view.View
import android.widget.FrameLayout
import android.widget.TextView
import androidx.appcompat.content.res.AppCompatResources
import androidx.constraintlayout.widget.ConstraintLayout
import androidx.core.view.ViewCompat
import androidx.recyclerview.widget.RecyclerView
import com.under9.android.comments.adapter.CommentItemClickListenerV2
import com.under9.android.comments.model.CommentItem
import com.under9.android.comments.model.constant.CommentConstant
import com.under9.android.comments.model.ui.CommentItemThemeAttr
import com.under9.android.comments.model.wrapper.CommentItemWrapperInterface
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.AVATAR_MODE_IMAGE
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_AVATAR_MODE
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_HIGHLIGHT_COMMENT_ID
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_INDENTATION_LEVEL
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_SHOULD_SHOW_COUNTRY
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_SHOULD_SHOW_LOAD_NEXT_CONTAINER
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_SHOULD_SHOW_LOAD_PREV_CONTAINER
import com.under9.android.comments.ui.renderer.CommentItemRendererDelegateV2.Companion.KEY_SHOULD_SHOW_OP
import com.under9.android.comments.ui.widget.CommentItemView
import com.under9.android.commentsystem.R
import com.under9.android.lib.network.UrlMapperInterface
import com.under9.android.lib.util.ColorUtils
import com.under9.android.lib.util.L10nUtil
import com.under9.android.lib.util.UiUtil
import com.under9.android.lib.widget.uiv3.UniversalImageView
import com.under9.android.lib.widget.uiv3.adapter.ImageTileInfoAdapter
import com.under9.android.lib.widget.uiv3.adapter.UIVAdapter
import com.under9.android.lib.widget.uiv3.adapter.VideoInfoAdapter
import com.under9.android.theme.ThemeStoreProvider
import kotlinx.android.synthetic.main.comment_system_list_item_inner_v5.view.*
import me.saket.bettermovementmethod.BetterLinkMovementMethod
import kotlin.math.pow

interface CommentItemInfoRendererInterface {
    fun bind(level: Int,
             wrapper: CommentItemWrapperInterface,
             themeAttr: CommentItemThemeAttr,
             viewHolder: RecyclerView.ViewHolder,
             position: Int)

    fun refreshRenderOptions(renderOptions: Bundle?)
}

abstract class BaseCommentItemRenderer(
        internal val commentItemClickListener: CommentItemClickListenerV2?,
        var renderOptions: Bundle?
): CommentItemInfoRendererInterface {

    override fun bind(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, position: Int) {
        bindCommentItemView(level, wrapper, themeAttr, viewHolder, viewHolder.itemView as CommentItemView, position)
    }

    internal fun bindTagAndListener(commentItemWrapper: CommentItemWrapperInterface,
                           view: View?, viewHolder: RecyclerView.ViewHolder, position: Int) {
        view?.apply {
            tag = commentItemWrapper
            setTag(R.id.commentPosition, position)
            setTag(R.id.commentItemViewHolder, viewHolder)
            setOnClickListener(commentItemClickListener)
        }
    }

    /**
     * After refreshing the list, call notifyDataSetChanged() to update the list rendering
     */
    override fun refreshRenderOptions(renderOptions: Bundle?) {
        this.renderOptions = renderOptions
    }

    abstract fun bindCommentItemView(level: Int,
                                     wrapper: CommentItemWrapperInterface,
                                     themeAttr: CommentItemThemeAttr,
                                     viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int)
}


class AvatarRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    private var dp16: Int = 0
    private var rootAvatarSize: Int = 0
    private var subLevelAvatarSize: Int = 0
    private var indentationLevel = 0
    private var avatarMode = 0

    init {
        refreshRenderOptions(renderOption)
    }

    override fun refreshRenderOptions(renderOptions: Bundle?) {
        super.refreshRenderOptions(renderOptions)
        indentationLevel = renderOptions?.getInt(KEY_INDENTATION_LEVEL, 0) ?: 0
        avatarMode = renderOptions?.getInt(KEY_AVATAR_MODE, AVATAR_MODE_IMAGE) ?: AVATAR_MODE_IMAGE
    }

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        var avatarUrl: String?
        with(commentItemView) {
            if (rootAvatarSize <= 0) {
                rootAvatarSize = resources.getDimensionPixelSize(R.dimen.avatar_size)
            }

            if (dp16 <= 0) {
                dp16 = UiUtil.dpToPx(context, 16)
            }

            val wrapperLevel = wrapper.getLevel() + indentationLevel - 1
            val avatarSize = (rootAvatarSize * 0.6.pow((wrapperLevel).toDouble())).toInt()
            val marginLeft = dp16 * (wrapperLevel)

            val layoutParams = avatar.layoutParams as ConstraintLayout.LayoutParams
            layoutParams.setMargins(marginLeft, layoutParams.topMargin, layoutParams.rightMargin, layoutParams.bottomMargin)
            layoutParams.width = avatarSize
            layoutParams.height = avatarSize

            val avatarLayoutParams = avatar.avatar.layoutParams as FrameLayout.LayoutParams
            avatarLayoutParams.width = avatarSize
            avatarLayoutParams.height = avatarSize

            val username = wrapper.getUser().getName()
            avatar.setTag(R.id.username, username)
            avatarUrl = wrapper.getUser().getAvatarUrl()

            when (avatarMode) {
                AVATAR_MODE_IMAGE -> avatar.setImageURI(if (avatarUrl == null) null else Uri.parse(avatarUrl))
                else -> setRoundAvatarColorDrawable(themeAttr.colorArray, themeAttr.textDrawableCache, themeAttr.avatarDrawable, username)
            }

            avatar.isActive = wrapper.getUser().isActive()

            bindTagAndListener(wrapper, avatar, viewHolder, position)
        }
    }
}

class UserInfoRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    private var shouldShowCountry: Boolean = false

    init {
        refreshRenderOptions(renderOption)
    }

    override fun refreshRenderOptions(renderOptions: Bundle?) {
        super.refreshRenderOptions(renderOptions)
        shouldShowCountry = renderOptions?.getBoolean(KEY_SHOULD_SHOW_COUNTRY, false) ?: false
    }

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        val user = wrapper.getUser()
        val username = user.getName()
        val emojiStatus = user.getEmojiStatus()
        val country = if (shouldShowCountry) user.getCountryEmoji() else ""

        with(commentItemView) {
            userName.text = if (emojiStatus.isEmpty() && country.isEmpty()) {
                username
            } else if (emojiStatus.isEmpty() && country.isNotEmpty()) {
                "$username $country"
            } else if (emojiStatus.isNotEmpty() && country.isEmpty()) {
                "$username $emojiStatus"
            } else {
                "$username $country $emojiStatus"
            }
            userName.setTag(R.id.username, username)

            // apply the pro badge if that user is pro user
            if (user.isPro()) {
                val isProPlus = user.isProPlus()
                proBadge.updateProBadge(isProPlus = isProPlus)
                proBadge.visibility = View.VISIBLE
            } else {
                proBadge.visibility = View.GONE
            }

            bindTagAndListener(wrapper, userName, viewHolder, position)
            bindTagAndListener(wrapper, proBadge, viewHolder, position)
        }
    }
}

class OpInfoRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {
    private var opUsernamePadding: Int = 0
    private var opUsernamePaddingTopBottom: Int = 0
    private var showOP = true

    init {
        refreshRenderOptions(renderOption)
    }

    override fun refreshRenderOptions(renderOptions: Bundle?) {
        super.refreshRenderOptions(renderOptions)
        showOP = renderOptions?.getBoolean(KEY_SHOULD_SHOW_OP) ?: true
    }

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        with(commentItemView) {
            if (opUsernamePadding <= 0) {
                opUsernamePadding = resources.getDimensionPixelSize(R.dimen.space4)
            }

            if (opUsernamePaddingTopBottom <= 0) {
                opUsernamePaddingTopBottom = resources.getDimensionPixelSize(R.dimen.cs_metaTextTopBottomPadding)
            }

            if (showOP && wrapper.isOP) {
                userName.setTextColor(themeAttr.getColor(R.styleable.CommentSystemItemView_cs_themeTextColorOpHighlighted))
                ViewCompat.setBackground(userName, themeAttr.opBackgroundDrawable)
                userName.setPadding(opUsernamePadding, opUsernamePaddingTopBottom, opUsernamePadding, opUsernamePaddingTopBottom)
            } else {
                // Don't highlight OP
                userName.setTextColor(themeAttr.getColor(R.attr.under9_themeTextColorPrimary))
                userName.setPadding(0, opUsernamePaddingTopBottom, 0, opUsernamePaddingTopBottom)
                ViewCompat.setBackground(userName, null)
            }
        }
    }
}

class TextContentRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    private var highlightCommentId: String? = null
    private var shouldUsernameClickable: Boolean = false
    private var shouldHighlightUrl: Boolean = false

    init {
        refreshRenderOptions(renderOption)
    }

    override fun refreshRenderOptions(renderOptions: Bundle?) {
        super.refreshRenderOptions(renderOptions)
        highlightCommentId = renderOptions?.getString(KEY_HIGHLIGHT_COMMENT_ID, null)
        shouldUsernameClickable = renderOptions?.getBoolean(CommentItemRendererDelegateV2.KEY_SHOULD_USERNAME_CLICKABLE, false) ?: false
        shouldHighlightUrl = renderOptions?.getBoolean(CommentItemRendererDelegateV2.KEY_SHOULD_HIGHLIGHT_URL, false) ?: false
    }

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        val textContent = wrapper.getContent()
        val content = commentItemView.content

        if (textContent.isNullOrEmpty()) {
            content.visibility = View.GONE
        } else {
            content.visibility = View.VISIBLE
        }

        val contentBuilder = SpannableStringBuilder(textContent)
        if (shouldUsernameClickable) {
            val matcher = CommentItemThemeAttr.mentionedUserPattern.matcher(textContent)
            while (matcher.find()) {
                val start = matcher.start()
                val end = matcher.end()
                val username = matcher.group()
                contentBuilder.setSpan(ForegroundColorSpan(themeAttr.getColor(R.attr.under9_themeColorAccent)), start, end, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                contentBuilder.setSpan(MentionedUserClickSpan(username, wrapper.getMentionedAccountId(username), commentItemClickListener), start, end, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        if (shouldHighlightUrl) {
            val matcher = CommentItemThemeAttr.urlPatternV2.matcher(textContent)
            while (matcher.find()) {
                val start = matcher.start()
                val end = matcher.end()
                val url = matcher.group()
                contentBuilder.setSpan(ForegroundColorSpan(themeAttr.getColor(R.attr.under9_themeColorAccent)), start, end, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
                contentBuilder.setSpan(UrlClickSpan(wrapper, url, commentItemClickListener), start, end, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        if (wrapper.shouldHighlight() || wrapper.getCommentId() == highlightCommentId) {
            content.setTextColor(themeAttr.getColor(R.attr.under9_themeTextColorPrimary))
        } else {
            content.setTextColor(themeAttr.getColor(R.attr.under9_themeTextColorPrimary))
        }

        content.text = contentBuilder
        content.movementMethod = BetterLinkMovementMethod.getInstance()
    }
}

class BodyRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    private var highlightCommentId: String? = null
    private var showLoadPrevContainer = false
    private var showLoadNextContainer = true

    private var highlightedColor = -1
    private var foregroundColor = -1

    init {
        refreshRenderOptions(renderOption)
    }

    override fun refreshRenderOptions(renderOptions: Bundle?) {
        super.refreshRenderOptions(renderOptions)
        highlightCommentId = renderOptions?.getString(KEY_HIGHLIGHT_COMMENT_ID, null)
        showLoadPrevContainer = renderOptions?.getBoolean(KEY_SHOULD_SHOW_LOAD_PREV_CONTAINER, false) ?: false
        showLoadNextContainer = renderOptions?.getBoolean(KEY_SHOULD_SHOW_LOAD_NEXT_CONTAINER, true) ?: true
    }

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        with(commentItemView) {
            commentItemView.progressBar.visibility = View.GONE

            if (highlightedColor == -1) {
                highlightedColor = UiUtil.themeAttributeToColor(R.attr.cs_themeForegroundHighlighted, context, -1)
            }

            if (foregroundColor == -1) {
                foregroundColor = UiUtil.themeAttributeToColor(R.attr.under9_themeForeground, context, -1)
            }

            if (showLoadNextContainer && level <= 1 && wrapper.getChildrenTotal() > 0) {
                commentItemView.loadMoreTxt.visibility = View.VISIBLE
                commentItemView.loadMoreTxt.text = L10nUtil.getFormattedQuantityString(context, R.plurals.view_n_replies, wrapper.getChildrenTotal())
            } else {
                commentItemView.loadMoreTxt.visibility = View.GONE
            }

            if (showLoadPrevContainer && wrapper.isParent() && wrapper.prevUrl != null && position == 0) {
                commentItemView.loadPrevTxt.visibility = View.VISIBLE
            } else {
                commentItemView.loadPrevTxt.visibility = View.GONE
            }

            if (wrapper.getChildrenTotal() > 0) {
                commentItemView.bottomPlaceholder.visibility = View.VISIBLE
            } else {
                commentItemView.bottomPlaceholder.visibility = View.GONE
            }

            this.setBackgroundColor(foregroundColor)
            commentBorder.setBackgroundColor(themeAttr.getColor(R.attr.under9_themeLineColor))

            when {
                wrapper.getCommentId() == highlightCommentId -> commentItemView.setBackgroundColor(UiUtil.themeAttributeToColor(R.attr.cs_themeForegroundHighlighted, context, -1))
                wrapper.shouldHighlight() -> {
                    val colorDrawbles = arrayOf(ColorDrawable(highlightedColor), ColorDrawable(foregroundColor))
                    val transitionDrawable = TransitionDrawable(colorDrawbles)
                    transitionDrawable.isCrossFadeEnabled = true
                    commentItemView.background = transitionDrawable
                    transitionDrawable.startTransition(5000)
                }
                else -> {
                    commentItemView.setBackgroundColor(foregroundColor)
                    meta.setTextColor(themeAttr.getColor(R.attr.under9_themeTextColorSecondary))
                }
            }

            if (wrapper.isPending()) {
                meta.text = "..."
            } else {
                meta.text = themeAttr.timeFormatter.formatTime(wrapper.getTime())
            }

            bindTagAndListener(wrapper, commentItemView, viewHolder, position)
            bindTagAndListener(wrapper, moreBtn, viewHolder, position)
            bindTagAndListener(wrapper, loadMoreTxt, viewHolder, position)
            bindTagAndListener(wrapper, loadPrevTxt, viewHolder, position)
            bindTagAndListener(wrapper, replyBtn, viewHolder, position)
        }
    }

}

class VoteInfoRenderer(
        commentItemClickListener: CommentItemClickListenerV2?,
        renderOption: Bundle?
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    private var accentColor = -1
    private var normalColor = -1

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        pureBind(wrapper, commentItemView)

        bindTagAndListener(wrapper, commentItemView.likeBtn, viewHolder, position)
        bindTagAndListener(wrapper, commentItemView.dislikeBtn, viewHolder, position)
    }

    fun pureBind(wrapper: CommentItemWrapperInterface, commentItemView: CommentItemView) {
        with(commentItemView) {

            if (accentColor == -1) {
                accentColor = com.under9.android.lib.util.UiUtil.themeAttributeToColor(com.under9.android.commentsystem.R.attr.under9_themeColorAccent, context, -1)
            }

            if (normalColor == -1) {
                normalColor = com.under9.android.lib.util.UiUtil.themeAttributeToColor(com.under9.android.commentsystem.R.attr.under9_themeTextColorSecondary, context, -1)
            }

            when (wrapper.likeStatus) {
                com.under9.android.comments.model.CommentItem.LIKE_STATUS_LIKED -> {
                    likeBtn.isChecked = true
                    dislikeBtn.isChecked = false

                    likeBtn.setTextColor(accentColor)
                    likeBtnMask.setTextColor(accentColor)
                    dislikeBtn.setTextColor(normalColor)
                    dislikeBtnMask.setTextColor(normalColor)
                }

                com.under9.android.comments.model.CommentItem.LIKE_STATUS_NORMAL -> {
                    likeBtn.isChecked = false
                    dislikeBtn.isChecked = false

                    likeBtn.setTextColor(normalColor)
                    likeBtnMask.setTextColor(normalColor)
                    dislikeBtn.setTextColor(normalColor)
                    dislikeBtnMask.setTextColor(normalColor)
                }

                com.under9.android.comments.model.CommentItem.LIKE_STATUS_DISLIKED -> {
                    likeBtn.isChecked = false
                    dislikeBtn.isChecked = true

                    likeBtn.setTextColor(normalColor)
                    likeBtnMask.setTextColor(normalColor)
                    dislikeBtn.setTextColor(accentColor)
                    dislikeBtnMask.setTextColor(accentColor)
                }
            }


            setVoteCount(wrapper.likeCount, likeBtn)
            setVoteCount(wrapper.dislikeCount, dislikeBtn)

            likeBtn.buttonDrawable = androidx.appcompat.content.res.AppCompatResources.getDrawable(context, com.under9.android.commentsystem.R.drawable.ic_comment_upvote_v2)
            dislikeBtn.buttonDrawable = androidx.appcompat.content.res.AppCompatResources.getDrawable(context, com.under9.android.commentsystem.R.drawable.ic_comment_downvote_v2)
            if (context is ThemeStoreProvider) {
                likeBtn.buttonDrawable = com.under9.android.lib.util.ColorUtils.getTintedVectorDrawable(context,
                        com.under9.android.commentsystem.R.drawable.ic_comment_upvote_v2,
                        (context as ThemeStoreProvider).getThemeStore().checkedColorStateList)

                dislikeBtn.buttonDrawable = com.under9.android.lib.util.ColorUtils.getTintedVectorDrawable(context,
                        com.under9.android.commentsystem.R.drawable.ic_comment_downvote_v2,
                        (context as ThemeStoreProvider).getThemeStore().checkedColorStateList)
            }
        }
    }

    private fun setVoteCount(count: Int, textView: TextView) {
        if (count == 0) {
            textView.text = ""
        } else {
            textView.text = L10nUtil.getFormatNumber(count)
        }
    }

}

class MediaRenderer(
        commentItemClickListener: CommentItemClickListenerV2,
        renderOption: Bundle?,
        private val urlMapperInterface: UrlMapperInterface
): BaseCommentItemRenderer(commentItemClickListener, renderOption) {

    override fun bindCommentItemView(level: Int, wrapper: CommentItemWrapperInterface, themeAttr: CommentItemThemeAttr, viewHolder: RecyclerView.ViewHolder, commentItemView: CommentItemView, position: Int) {
        val existingWrapper = commentItemView.uiv.getTag(R.id.comment_wrapper)

        if (wrapper.getType() != CommentConstant.TYPE_MEDIA) {
            commentItemView.uiv.visibility = View.GONE
            return
        } else {
            commentItemView.uiv.visibility = View.VISIBLE
        }

        if (existingWrapper != null && existingWrapper == wrapper) return

        val imageMediaByType = wrapper.getFirstMedia()?.imageMetaByType
        
        when (imageMediaByType?.type ?: return) {
            CommentConstant.MEDIA_TYPE_STATIC -> {
                val embedImage = imageMediaByType.image
                val height = embedImage.height
                val width = embedImage.width
                val url = embedImage.getImageUrl()

                val imageTileInfoAdapter = ImageTileInfoAdapter.newBuilder()
                        .setImage(url, width, height)
                        .build()

                val adapter = UIVAdapter.newBuilder(UniversalImageView.MODE_STATIC_IMAGE)
                        .setUrlMapper(urlMapperInterface)
                        .setRetryDrawable(R.drawable.icn_retry)
                        .setImageTileInfoAdapter(imageTileInfoAdapter)
                        .setOnClickListener(commentItemClickListener)
                        .setOnLongClickListener(commentItemClickListener)
                        .build()
                commentItemView.uiv.adapter = adapter
            }

            CommentConstant.MEDIA_TYPE_ANIMATED -> {
                val video = imageMediaByType.video
                val image = imageMediaByType.image

                if (video != null && image != null) {
                    val url = video.url

                    val height = image.height

                    val videoInfoAdapter = VideoInfoAdapter.newBuilder()
                            .videoUrl(url)
                            .build()

                    val imageTileInfoAdapter = ImageTileInfoAdapter.newBuilder()
                            .setImage(image.getImageUrl(), image.width, height)
                            .build()

                    val adapter = UIVAdapter.newBuilder(UniversalImageView.MODE_MP4_WITHOUT_SOUND)
                            .setRetryDrawable(R.drawable.icn_retry)
                            .centerBadge(R.drawable.ic_play_gif)
                            .setUrlMapper(urlMapperInterface)
                            .setImageTileInfoAdapter(imageTileInfoAdapter)
                            .setVideoInfoAdapter(videoInfoAdapter)
                            .setOnLongClickListener(commentItemClickListener)
                            .build()

                    commentItemView.uiv.adapter = adapter
                } else {
                    return
                }
            }
        }

        bindTagAndListener(wrapper, commentItemView.refresh, viewHolder, position)
        commentItemView.uiv.apply {
            setTag(R.id.commentPosition, position)
            setTag(R.id.comment_wrapper, wrapper)
            setTag(R.id.image_holder, viewHolder)
        }
    }
}