package com.under9.android.lib.bottomsheet

import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import androidx.core.content.ContextCompat
import androidx.recyclerview.widget.RecyclerView
import com.under9.android.lib.bottomsheet.BottomSheet.TYPE_ACCOUNT
import com.under9.android.lib.bottomsheet.BottomSheet.TYPE_BADGE
import com.under9.android.lib.bottomsheet.BottomSheet.TYPE_DIVIDER
import com.under9.android.lib.bottomsheet.BottomSheet.TYPE_SIMPLE
import com.under9.android.lib.bottomsheet.BottomSheet.TYPE_SWITCH
import com.under9.android.lib.util.UiUtil
import com.under9.android.lib.widget.R
import kotlinx.android.synthetic.main.view_item_bottom_sheet_account.view.*
import kotlinx.android.synthetic.main.view_item_bottom_sheet_switch.view.*
import kotlinx.android.synthetic.main.view_item_simple_bottom_sheet.view.menuIcon
import kotlinx.android.synthetic.main.view_item_simple_bottom_sheet.view.menuTitle


class BottomSheetAdapterV2(private val items: List<BottomSheetModel>,
                           private val clickListener: BottomSheetClickListener,
                           private val dismissListener: DismissListener) : RecyclerView.Adapter<RecyclerView.ViewHolder>() {


    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int) = when (viewType) {
        TYPE_ACCOUNT -> AccountItemView(LayoutInflater.from(parent.context).inflate(R.layout.view_item_bottom_sheet_account, parent, false), clickListener, dismissListener)
        TYPE_SWITCH -> SwitchItemView(LayoutInflater.from(parent.context).inflate(R.layout.view_item_bottom_sheet_switch, parent, false), clickListener, dismissListener)
        TYPE_BADGE -> BadgeItemView(LayoutInflater.from(parent.context).inflate(R.layout.view_item_bottom_sheet_badge, parent, false), clickListener, dismissListener)
        TYPE_SIMPLE -> SimpleViewHolder(LayoutInflater.from(parent.context).inflate(R.layout.view_item_simple_bottom_sheet, parent, false), clickListener, dismissListener)
        TYPE_DIVIDER -> DividerItemView(LayoutInflater.from(parent.context).inflate(R.layout.view_item_bottom_sheet_divider, parent, false))
        else -> SimpleViewHolder(LayoutInflater.from(parent.context).inflate(R.layout.view_item_simple_bottom_sheet, parent, false), clickListener, dismissListener)
    }

    override fun getItemCount() = items.size

    override fun onBindViewHolder(holder: RecyclerView.ViewHolder, position: Int) {
        when (getItemViewType(position)) {
            TYPE_ACCOUNT -> (holder as AccountItemView).bind(items[position].title, items[position].additionalStringInfo!!, position, items[position].additionalBooleanInfo, items[position].id)
            TYPE_SWITCH -> (holder as SwitchItemView).bind(items[position].title, items[position].iconRes, position, items[position].additionalBooleanInfo!!, items[position].id, additionalIconColor = items[position].additionalIconColor)
            TYPE_BADGE -> (holder as BadgeItemView).bind(items[position].title, items[position].iconRes, position, items[position].additionalBooleanInfo, items[position].id)
            TYPE_SIMPLE -> (holder as SimpleViewHolder).bind(items[position].title, items[position].iconRes, position, items[position].id, additionalIconColor = items[position].additionalIconColor)
            TYPE_DIVIDER -> return
            else -> (holder as SimpleViewHolder).bind(items[position].title, items[position].iconRes, position, items[position].id, additionalIconColor = items[position].additionalIconColor)
        }
        holder.itemView.tag = items[position].id
    }


    override fun getItemViewType(position: Int) = items[position].displayType

    open class SimpleViewHolder(itemView: View,
                                private val clickListener: BottomSheetClickListener,
                                private val dismissListener: DismissListener) : RecyclerView.ViewHolder(itemView) {

        fun bind(title: String, resIcon: Int, pos: Position, id: MenuItemId, dismissible: Boolean = true, additionalIconColor: Int) {
            with(itemView) {
                menuTitle.text = title

                setBackgroundColor(UiUtil.themeAttributeToColor(R.attr.under9_themeForeground, context, -1))
                menuTitle.setTextColor(UiUtil.themeAttributeToColor(R.attr.under9_themeTextColorPrimary, context, -1))

                menuIcon.setImageDrawable(ContextCompat.getDrawable(itemView.context, resIcon))
                if (additionalIconColor == -1) {
                    menuIcon.applyTintColor(UiUtil.themeAttributeToColor(R.attr.under9_themeIconColor, context, -1))
                } else {
                    menuIcon.applyTintColor(ContextCompat.getColor(context, additionalIconColor))
                }
                setOnClickListener {
                    clickListener(pos, id)
                    if (dismissible) dismissListener()
                }


            }
        }
    }


    class AccountItemView(itemView: View,
                          private val clickListener: BottomSheetClickListener,
                          private val dismissListener: DismissListener) : RecyclerView.ViewHolder(itemView) {

        fun bind(username: String, avatarUri: String, pos: Position, isPro: Boolean?, id: MenuItemId) {
            with(itemView) {
                accountAvatar.setImageURI(avatarUri)
                accountAvatar.isActive = avatarUri.isNotBlank()
                tvUsername.text = username
                tvUsername.setTextColor(UiUtil.themeAttributeToColor(R.attr.under9_themeTextColorPrimary, context, -1))

                setBackgroundColor(UiUtil.themeAttributeToColor(R.attr.under9_themeForeground, context, -1))
                isPro?.let {
                    proBadge.visibility = View.VISIBLE
                    proBadge.updateProBadge(it)
                } ?: run {
                    proBadge.visibility = View.GONE
                }

                setOnClickListener {
                    clickListener(pos, id)
                    dismissListener()
                }
            }
        }
    }

    class BadgeItemView(itemView: View,
                        clickListener: BottomSheetClickListener,
                        dismissListener: DismissListener) : SimpleViewHolder(itemView, clickListener, dismissListener) {

        fun bind(title: String, resIcon: Int, pos: Position, isPro: Boolean?, id: MenuItemId) {
            super.bind(title, resIcon, pos, id, true, -1)
            isPro?.let {
                itemView.proBadge.visibility = View.VISIBLE
                itemView.proBadge.updateProBadge(it)
            } ?: run {
                itemView.proBadge.visibility = View.GONE
            }
        }
    }

    class SwitchItemView(itemView: View,
                         clickListener: BottomSheetClickListener,
                         dismissListener: DismissListener) : SimpleViewHolder(itemView, clickListener, dismissListener) {

        fun bind(title: String, resIcon: Int, pos: Position, isChecked: Boolean, id: MenuItemId, additionalIconColor: Int) {
            super.bind(title, resIcon, pos, id, false, additionalIconColor = additionalIconColor)
            itemView.actionSwitch.isChecked = isChecked
        }
    }

    class DividerItemView(itemView: View) : RecyclerView.ViewHolder(itemView)
}